import datetime as DT
import itertools

from libestg3b import EstG3b
from libestg3b.matcher import Match


def _matchers(e, *descriptions):
    found = set()
    descriptions = set(descriptions)

    for matcher in itertools.chain.from_iterable(e._matchers):
        if matcher._description in descriptions:
            found.add(matcher)
            descriptions.remove(matcher._description)

    if descriptions:
        raise LookupError('Could not find ' + ','.join(descriptions))

    return found


def test_estg3b():
    e = EstG3b('DE')
    match = e.calculate_shift([DT.datetime(2018, 2, 1, 2), DT.datetime(2018, 2, 1, 6)])
    assert isinstance(match, list)
    assert len(match) == 1
    assert match[0] == Match(DT.datetime(2018, 2, 1, 2), DT.datetime(2018, 2, 1, 6), _matchers(e, 'Nachtarbeit 20:00-06:00'))


def test_estg3b_multimatch():
    e = EstG3b('DE')
    match = e.calculate_shift([DT.datetime(2018, 2, 1, 2), DT.datetime(2018, 2, 1, 7)])
    assert isinstance(match, list)
    assert len(match) == 2
    assert match[0] == Match(DT.datetime(2018, 2, 1, 2), DT.datetime(2018, 2, 1, 6), _matchers(e, 'Nachtarbeit 20:00-06:00'))
    assert match[1] == Match(DT.datetime(2018, 2, 1, 6), DT.datetime(2018, 2, 1, 7), set())


def test_estg3b_nomatch():
    e = EstG3b('DE')
    match = e.calculate_shift([DT.datetime(2018, 2, 1, 8), DT.datetime(2018, 2, 1, 9)])
    assert isinstance(match, list)
    assert len(match) == 1
    assert match[0] == Match(DT.datetime(2018, 2, 1, 8), DT.datetime(2018, 2, 1, 9), set())


def test_estg3b_sunday_plus_night():
    e = EstG3b('DE')
    match = e.calculate_shift([DT.datetime(2018, 9, 16, 20), DT.datetime(2018, 9, 16, 22)])
    assert isinstance(match, list)
    assert len(match) == 1
    assert match[0] == Match(DT.datetime(2018, 9, 16, 20), DT.datetime(2018, 9, 16, 22), _matchers(e, 'Nachtarbeit 20:00-06:00', 'Sonntagsarbeit'))



import datetime
import inspect
import itertools
from decimal import Decimal

import holidays
from dateutil.relativedelta import relativedelta

from .matcher import DayMatcher, DayTimeMatcher, Match
from .matcher import Matcher as M


def EstG3b(country):
    try:
        country_estg3b = globals()['EstG3b' + country]()
    except (KeyError):
        raise KeyError("Country %s not available" % country)
    return country_estg3b


def EstG3bs():
    return [clazz for clazz in globals().values() if inspect.isclass(clazz) and issubclass(clazz, EstG3bBase) and clazz != EstG3bBase]


class EstG3bBase:
    def __init__(self, country, base_matchers, add_matchers=None, replace_matchers=None) -> None:
        self._holidays = holidays.CountryHoliday(country.upper())
        self._matchers = base_matchers

        if replace_matchers:
            self._matchers = replace_matchers.copy()
        if add_matchers:
            self._machers.extend(add_matchers)

        assert self._matchers

    def _list_minutes(self, start, end):
        start = start.replace(second=0, microsecond=0)
        end = end.replace(second=0, microsecond=0)

        while start < end:
            yield start
            start = start + relativedelta(minutes=1)

    def calculate_shift(self, shift):
        assert len(shift) == 2
        assert isinstance(shift[0], datetime.datetime)
        assert isinstance(shift[1], datetime.datetime)

        minutes = self._list_minutes(shift[0], shift[1])
        first_minute = next(minutes)
        minutes = itertools.chain([first_minute], minutes)

        matches = []

        for minute in minutes:
            minute_matchers = []

            # find the highest matcher for each group (if any) and save it
            for group in self._matchers:
                try:
                    minute_matchers.append(max([
                        matcher for matcher in group
                        if matcher(minute, first_minute, self._holidays)
                    ]))
                except ValueError:  # no match found
                    pass

            minute_matchers = set(minute_matchers)

            if matches and matches[-1].matchers == minute_matchers:
                # combine equal matches by increasing the length of the last one
                matches[-1].end = matches[-1].end + relativedelta(minutes=1)
            else:
                # a list of minutes is inclusive the last one, the `end` stamp is exclusive
                matches.append(Match(minute, minute + relativedelta(minutes=1), minute_matchers))

        return matches

    def calculate_shifts(self, shifts):
        return map(self.calculate_shift, shifts)


class EstG3bGermany(EstG3bBase):
    def __init__(self, **kwargs):
        matchers = (
            (
                M(
                    'Nachtarbeit 20:00-06:00', lambda m: m.hour >= 20 or m.hour < 6, multiply=Decimal('0.25'),
                    tests=(
                        ('~19:59', False),
                        ('~20:00', True),
                        ('~21:00', True),
                        ('~05:59', True),
                        ('~06:00', False),
                    ),
                ),
                M(
                    'Nachtarbeit 00:00-04:00 (Folgetag)', lambda m, f: 0 <= m.hour < 4 and f.date() < m.date(), multiply=Decimal('0.4'),
                    tests=(
                        ('00:00~00:01', False),
                        ('23:59~00:01', True),
                        ('23:59~03:59', True),
                        ('23:59~04:00', False),
                    )
                ),
            ),
            (
                M(
                    'Sonntagsarbeit', lambda m: m.weekday() == 6, multiply=Decimal('0.5'),
                    tests=(
                        ('~2018-09-15', False),
                        ('~2018-09-16 00:00', True),
                        ('~2018-09-16 23:59', True),
                        ('2018-09-16 23:59~2018-09-17 00:00', False),
                    ),
                ),
                M(
                    'Sonntagsarbeit (Montag)', lambda m, f: f.weekday() == 6 and 0 <= m.hour < 4, multiply=Decimal('0.5'),
                    tests=(
                        # 2018-09-16 is a Sunday
                        ('~2018-09-16 00:00', False),
                        ('~2018-09-16 23:59', False),
                        ('2018-09-16 23:59~2018-09-17 00:00', True),
                        ('2018-09-16 23:59~2018-09-17 03:59', True),
                        ('2018-09-16 23:59~2018-09-17 04:00', False),
                        ('2018-09-17 23:59~2018-09-18 00:00', False),
                    ),
                ),
                M(
                    'Feiertagsarbeit', lambda m, f, holidays: m in holidays, multiply=Decimal('1.25'),
                    tests=(
                        # 2018-05-10 is Christi Himmelfahrt
                        ('~2018-05-09 23:59', False),
                        ('~2018-05-10 00:00', True),
                        ('~2018-05-10 23:59', True),
                        ('~2018-05-11 00:00', False),
                        ('2018-05-10 23:59~2018-05-11 00:00', False),
                    ),
                ),
                M(
                    'Feiertagsarbeit (Folgetag)', lambda m, f, holidays: f in holidays and 0 <= m.hour < 4, multiply=Decimal('1.25'),
                    tests=(
                        # 2018-05-10 is Christi Himmelfahrt
                        ('~2018-05-10 00:00', False),
                        ('~2018-05-10 23:59', False),
                        ('2018-05-10 23:59~2018-05-11 00:00', True),
                        ('2018-05-10 23:59~2018-05-11 03:59', True),
                        ('2018-05-10 23:59~2018-05-11 04:00', False),
                        ('2018-09-11 23:59~2018-05-12 00:00', False),
                    )
                ),
                DayTimeMatcher(12, 24, 14, multiply=Decimal('1.25')),
                DayTimeMatcher(12, 31, 14, multiply=Decimal('1.25')),
                DayMatcher(12, 25, multiply=Decimal('1.5')),
                DayMatcher(12, 26, multiply=Decimal('1.5')),
                DayMatcher(5, 1, multiply=Decimal('1.5')),
            ),
        )
        super().__init__('DE', matchers, **kwargs)


EstG3bDE = EstG3bGermany



import itertools

import pytest

from libestg3b.estg3b import EstG3bs


def _generate_matcher_examples(e):
    for m in itertools.chain.from_iterable(e._matchers):  # flatten groups
        yield from ([m, e._holidays, *ex] for ex in m.examples())


@pytest.mark.parametrize(
    'matcher,holidays,minute,first_minute,result',
    itertools.chain.from_iterable(_generate_matcher_examples(e()) for e in EstG3bs())
)
def test_matchers(matcher, minute, first_minute, holidays, result):
    assert matcher(minute, first_minute, holidays) == result



import datetime
from decimal import Decimal

import pytest

from libestg3b.matcher import DayMatcher, DayTimeMatcher, Matcher


def test_matcher_init():
    m = Matcher('a', lambda x: False, add=1)
    m(None, None, None)


def test_matcher_init__no_add_or_multiply():
    with pytest.raises(AssertionError):
        Matcher('a', lambda x: False)


def test_matcher_init_add_and_multiply():
    with pytest.raises(AssertionError):
        Matcher('a', lambda x: False, add=1, multiply=2)


def test_matcher_init_no_description():
    with pytest.raises(AssertionError):
        Matcher('', lambda x: False)


@pytest.mark.parametrize("impl", [
    lambda: False,
    lambda a, b, c, d: False,
])
def test_matcher_init_impl_nargs_crash(impl):
    with pytest.raises(AssertionError):
        Matcher('a', impl)


@pytest.mark.parametrize("impl", [
    lambda a: False,
    lambda a: True,
])
def test_matcher_call_rtn_type(impl):
    m = Matcher('a', impl, add=1)
    m(None, None, None)


def test_matcher_call_rtn_type_crash():
    m = Matcher('a', lambda x: None, add=1)
    with pytest.raises(AssertionError):
        m(None, None, None)


def test_matcher_call_passing_1():
    args = None

    def impl(a):
        nonlocal args
        args = [a]
        return True

    m = Matcher('a', impl, add=1)
    m(0, 1, 2)
    assert args == [0]


def test_matcher_call_passing_2():
    args = None

    def impl(a, b):
        nonlocal args
        args = [a, b]
        return True

    m = Matcher('a', impl, add=1)
    m(0, 1, 2)
    assert args == [0, 1]


def test_matcher_call_passing_3():
    args = None

    def impl(a, b, c):
        nonlocal args
        args = [a, b, c]
        return True

    m = Matcher('a', impl, add=1)
    m(0, 1, 2)
    assert args == [0, 1, 2]


@pytest.mark.parametrize('dt,res', [
    (datetime.datetime(2018, 12, 25, 23, 59), False),
    (datetime.datetime(2018, 12, 26, 0), True),
    (datetime.datetime(2018, 12, 26, 23, 59), True),
    (datetime.datetime(2018, 12, 27, 0), False),
    (datetime.datetime(2019, 12, 26, 0), True),
])
def test_daymatcher(dt, res):
    m = DayMatcher(12, 26, multiply=Decimal('1.25'))
    assert m(dt, dt, None) == res


@pytest.mark.parametrize('dt,res', [
    (datetime.datetime(2018, 12, 24, 0), False),
    (datetime.datetime(2018, 12, 24, 13, 59), False),
    (datetime.datetime(2018, 12, 24, 14), True),
    (datetime.datetime(2019, 12, 24, 14), True),
    (datetime.datetime(2018, 12, 25, 14), False),
    (datetime.datetime(2018, 12, 24, 23, 59), True),
])
def test_daytimematcher(dt, res):
    m = DayTimeMatcher(12, 24, 14, multiply=Decimal('1.25'))
    assert m(dt, dt, None) == res



from __future__ import print_function
import sys

try:
    from setuptools import setup
except ImportError:
    print("libestg3b needs setuptools.", file=sys.stderr)
    print("Please install it using your package-manager or pip.", file=sys.stderr)
    sys.exit(1)

setup(
    name='libestg3b',
    version='0.0.1',
    description='',
    author='uberspace.de',
    author_email='hallo@uberspace.de',
    url='https://github.com/uberspace/libestg3b',
    packages=[
        'libestg3b',
    ],
    install_requires=[
        'holidays',
        'pytz',
        'python-dateutil',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Office/Business :: Financial',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    zip_safe=True,
)



import datetime
from decimal import Decimal
from inspect import signature
from typing import Optional, Set

import dateutil
from holidays import CountryHoliday

from dataclasses import dataclass


class Matcher():
    def __init__(self, description: str, impl, *, multiply: Optional[Decimal] = None, add: Optional[Decimal] = None, tests=[]) -> None:
        self._description = description
        self._impl = impl
        self._multiply = multiply
        self._add = add
        self._tests = tests

        assert (multiply is None) != (add is None)
        assert len(description) > 0
        assert 1 <= len(self._impl_parameters) <= 3

    @property
    def _impl_parameters(self):
        return signature(self._impl).parameters

    @property
    def _bonus(self):
        return self._multiply if self._multiply else self._add

    def _parse_test_time(self, tt):
        default = datetime.datetime(2018, 1, 10)
        return dateutil.parser.parse(tt, default=default) if tt else default

    def examples(self):
        for t in self._tests:
            first_minute, minute = t[0].split('~')
            minute = self._parse_test_time(minute)
            first_minute = self._parse_test_time(first_minute)

            if minute.time() < first_minute.time():
                minute = minute + dateutil.relativedelta.relativedelta(days=1)

            yield [
                minute,
                first_minute,
                t[1],
            ]

    def __call__(self, minute: datetime.datetime, first_minute: datetime.datetime, holidays: CountryHoliday) -> bool:
        narg = len(self._impl_parameters)

        if narg == 1:
            r = self._impl(minute)
        elif narg == 2:
            r = self._impl(minute, first_minute)
        elif narg == 3:
            r = self._impl(minute, first_minute, holidays)

        assert isinstance(r, bool)

        return r

    def __repr__(self):
        return f'<Matcher: {self._description}>'

    def __hash__(self):
        return hash(self._description)

    def __eq__(self, other):
        return self._description == other._description

    def __gt__(self, other):
        return self._bonus > other._bonus

    def __lt__(self, other):
        return self._bonus < other._bonus


class DayMatcher(Matcher):
    """ match, if the given minute is within the given day. Keyword arguments are passed onto Matcher. """
    def __init__(self, month: int, day: int, **kwargs) -> None:
        super().__init__(
            f'{month:02d}-{day:02d}',
            lambda m: m.month == month and m.day == day,
            **kwargs,
        )


class DayTimeMatcher(Matcher):
    """ match, if the given minute is within the given day after the given hour. Keyword arguments are passed onto Matcher. """
    def __init__(self, month: int, day: int, hour: int, **kwargs) -> None:
        super().__init__(
            f'{month:02d}-{day:02d} {hour:02d}:00+',
            lambda m: m.month == month and m.day == day and m.hour >= hour,
            **kwargs,
        )


@dataclass
class Match():
    start: datetime.datetime
    end: datetime.datetime
    matchers: Set[Matcher]
